unit NMD5;

interface

type
  TMD5 = class(TObject)
  private
    State: array[0..3] of Longword;
    Count: array[0..1] of Longword;
    Buffer: array[0..63] of Byte;
  public
    constructor Create;
    procedure Update(const Buf; NBytes: Cardinal);
    function Digest: string;
    function HexDigest: string;
  end;

implementation

procedure FF(var A: Longword; B, C, D, X, S, Ac: Longword);
begin
  A := A + ((B and C) or ((not B) and D)) + X + Ac;
  A := (A shl S) or (A shr (32 - S)); // rotate left
  Inc(A, B);
end;

procedure GG(var A: Longword; B, C, D, X, S, Ac: Longword);
begin
  A := A + ((B and D) or (C and (not D))) + X + Ac;
  A := (A shl S) or (A shr (32 - S)); // rotate left
  Inc(A, B);
end;

procedure HH(var A: Longword; B, C, D, X, S, Ac: Longword);
begin
  A := A + (B xor C xor D) + X + Ac;
  A := (A shl S) or (A shr (32 - S)); // rotate left
  Inc(A, B);
end;

procedure II(var A: Longword; B, C, D, X, S, Ac: Longword);
begin
  A := A + (C xor (B or (not D))) + X + Ac;
  A := (A shl S) or (A shr (32 - S)); // rotate left
  Inc(A, B);
end;

procedure EncodeLE(const InBuf: array of Longword; var OutBuf);
var
  i: Cardinal;
begin
  for i := 0 to High(InBuf) do
  begin
    PChar(@OutBuf)[i*4] := Char(InBuf[i] and $ff);
    PChar(@OutBuf)[i*4+1] := Char(InBuf[i] shr 8 and $ff);
    PChar(@OutBuf)[i*4+2] := Char(InBuf[i] shr 16 and $ff);
    PChar(@OutBuf)[i*4+3] := Char(InBuf[i] shr 24 and $ff);
  end;
end;

procedure DecodeLE(const InBuf; InBytes: Cardinal;
  var OutBuf: array of Longword);
var
  i: Cardinal;
begin
  for i := 0 to (InBytes div 4)-1 do
    OutBuf[i] := Longword(PChar(@InBuf)[i*4]) or
                 Longword(PChar(@InBuf)[i*4+1]) shl 8 or
                 Longword(PChar(@InBuf)[i*4+2]) shl 16 or
                 Longword(PChar(@InBuf)[i*4+3]) shl 24;
end;

procedure Transform(var State: array of Longword; const Block);
const
  s11 = 7;
  s12 = 12;
  s13 = 17;
  s14 = 22;
  s21 = 5;
  s22 = 9;
  s23 = 14;
  s24 = 20;
  s31 = 4;
  s32 = 11;
  s33 = 16;
  s34 = 23;
  s41 = 6;
  s42 = 10;
  s43 = 15;
  s44 = 21;
var
  a, b, c, d: Longword;
  x: array[0..15] of Longword;
begin
  DecodeLE(Block, 64, x);

  a := State[0];
  b := State[1];
  c := State[2];
  d := State[3];

  // Round 1
  FF(a, b, c, d, x[ 0], s11, $d76aa478);
  FF(d, a, b, c, x[ 1], s12, $e8c7b756);
  FF(c, d, a, b, x[ 2], s13, $242070db);
  FF(b, c, d, a, x[ 3], s14, $c1bdceee);
  FF(a, b, c, d, x[ 4], s11, $f57c0faf);
  FF(d, a, b, c, x[ 5], s12, $4787c62a);
  FF(c, d, a, b, x[ 6], s13, $a8304613);
  FF(b, c, d, a, x[ 7], s14, $fd469501);
  FF(a, b, c, d, x[ 8], s11, $698098d8);
  FF(d, a, b, c, x[ 9], s12, $8b44f7af);
  FF(c, d, a, b, x[10], s13, $ffff5bb1);
  FF(b, c, d, a, x[11], s14, $895cd7be);
  FF(a, b, c, d, x[12], s11, $6b901122);
  FF(d, a, b, c, x[13], s12, $fd987193);
  FF(c, d, a, b, x[14], s13, $a679438e);
  FF(b, c, d, a, x[15], s14, $49b40821);

  // Round 2
  GG(a, b, c, d, x[ 1], s21, $f61e2562);
  GG(d, a, b, c, x[ 6], s22, $c040b340);
  GG(c, d, a, b, x[11], s23, $265e5a51);
  GG(b, c, d, a, x[ 0], s24, $e9b6c7aa);
  GG(a, b, c, d, x[ 5], s21, $d62f105d);
  GG(d, a, b, c, x[10], s22,  $2441453);
  GG(c, d, a, b, x[15], s23, $d8a1e681);
  GG(b, c, d, a, x[ 4], s24, $e7d3fbc8);
  GG(a, b, c, d, x[ 9], s21, $21e1cde6);
  GG(d, a, b, c, x[14], s22, $c33707d6);
  GG(c, d, a, b, x[ 3], s23, $f4d50d87);
  GG(b, c, d, a, x[ 8], s24, $455a14ed);
  GG(a, b, c, d, x[13], s21, $a9e3e905);
  GG(d, a, b, c, x[ 2], s22, $fcefa3f8);
  GG(c, d, a, b, x[ 7], s23, $676f02d9);
  GG(b, c, d, a, x[12], s24, $8d2a4c8a);

  // Round 3
  HH(a, b, c, d, x[ 5], s31, $fffa3942);
  HH(d, a, b, c, x[ 8], s32, $8771f681);
  HH(c, d, a, b, x[11], s33, $6d9d6122);
  HH(b, c, d, a, x[14], s34, $fde5380c);
  HH(a, b, c, d, x[ 1], s31, $a4beea44);
  HH(d, a, b, c, x[ 4], s32, $4bdecfa9);
  HH(c, d, a, b, x[ 7], s33, $f6bb4b60);
  HH(b, c, d, a, x[10], s34, $bebfbc70);
  HH(a, b, c, d, x[13], s31, $289b7ec6);
  HH(d, a, b, c, x[ 0], s32, $eaa127fa);
  HH(c, d, a, b, x[ 3], s33, $d4ef3085);
  HH(b, c, d, a, x[ 6], s34,  $4881d05);
  HH(a, b, c, d, x[ 9], s31, $d9d4d039);
  HH(d, a, b, c, x[12], s32, $e6db99e5);
  HH(c, d, a, b, x[15], s33, $1fa27cf8);
  HH(b, c, d, a, x[ 2], s34, $c4ac5665);

  // Round 4
  II(a, b, c, d, x[ 0], s41, $f4292244);
  II(d, a, b, c, x[ 7], s42, $432aff97);
  II(c, d, a, b, x[14], s43, $ab9423a7);
  II(b, c, d, a, x[ 5], s44, $fc93a039);
  II(a, b, c, d, x[12], s41, $655b59c3);
  II(d, a, b, c, x[ 3], s42, $8f0ccc92);
  II(c, d, a, b, x[10], s43, $ffeff47d);
  II(b, c, d, a, x[ 1], s44, $85845dd1);
  II(a, b, c, d, x[ 8], s41, $6fa87e4f);
  II(d, a, b, c, x[15], s42, $fe2ce6e0);
  II(c, d, a, b, x[ 6], s43, $a3014314);
  II(b, c, d, a, x[13], s44, $4e0811a1);
  II(a, b, c, d, x[ 4], s41, $f7537e82);
  II(d, a, b, c, x[11], s42, $bd3af235);
  II(c, d, a, b, x[ 2], s43, $2ad7d2bb);
  II(b, c, d, a, x[ 9], s44, $eb86d391);

  Inc(State[0], a);
  Inc(State[1], b);
  Inc(State[2], c);
  Inc(State[3], d);
end;

constructor TMD5.Create;
begin
  State[0] := $67452301;
  State[1] := $efcdab89;
  State[2] := $98badcfe;
  State[3] := $10325476;
end;

procedure TMD5.Update(const Buf; NBytes: Cardinal);
var
  i, index, partlen: Cardinal;
begin
  index := (Count[0] shr 3) and $3F;
  Inc(Count[0], NBytes shl 3);
  if Count[0] < (NBytes shl 3) then
    Inc(Count[1]);
  Inc(Count[1], NBytes shr 29);

  partlen := 64 - index;
  if NBytes >= partlen then
  begin
    Move(Buf, Buffer[index], partlen);
    Transform(State, Buffer);
    i := partlen;
    while (i + 63) < NBytes do
    begin
      Transform(State, PChar(@Buf)[i]);
      Inc(i, 64);
    end;
    index := 0;
  end else
    i := 0;
  Move(PChar(@Buf)[i], Buffer[index], NBytes-i);
end;

function TMD5.Digest: string;
const
  padding: array[0..63] of Byte = (
    $80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
  );
var
  bits: array[0..7] of Char;
  index, padlen: Cardinal;
  state: array[0..3] of Longword;
  count: array[0..1] of Longword;
begin
  Move(Self.State, state, SizeOf(state));
  Move(Self.Count, count, SizeOf(count));
  EncodeLE(count, bits);
  index := (count[0] shr 3) and $3f;
  if index < 56 then
    padlen := (56 - index)
  else
    padlen := (120 - index);
  Update(padding, padlen);
  Update(bits, 8);
  SetString(Result, nil, 16);
  EncodeLE(Self.State, Result[1]);
  Move(state, Self.State, SizeOf(Self.State));
  Move(count, Self.Count, SizeOf(Self.Count));
end;

function TMD5.HexDigest: string;
const
  hexdigs: array[0..15] of Char = '0123456789abcdef';
var
  i: Integer;
  s: string;
begin
  s := Digest;
  SetString(Result, nil, Length(s)*2);
  for i := 1 to Length(s) do
  begin
    Result[(i-1)*2+1] := hexdigs[Byte(s[i]) shr 4];
    Result[(i-1)*2+2] := hexdigs[Byte(s[i]) and $F];
  end;
end;

end.
